import copy
import json
from unittest import mock

from openstackclient.tests.unit import fakes, utils

from doniclient.osc import cli as hardware_cli
from doniclient.tests.osc import fakes as hardware_fakes


class TestHardware(hardware_fakes.TestHardware):
    def setUp(self):
        super(TestHardware, self).setUp()

        # Get a shortcut to the baremetal manager mock
        self.hardware_mock = self.app.client_manager.inventory
        self.hardware_mock.reset_mock()


class TestHardwareShow(TestHardware):
    def setUp(self):
        super().setUp()

        self.hardware_mock.get_by_uuid.return_value = (
            hardware_fakes.FakeHardware.create_one_hardware()
        )

        self.cmd = hardware_cli.GetHardware(self.app, None)

    def test_hardware_show(self):
        arglist = ["xxx-xxxxxx-xxxx"]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = ["xxx-xxxxxx-xxxx"]

        self.hardware_mock.get_by_uuid.assert_called_with(*args)

        collist = (
            "created_at",
            "hardware_type",
            "name",
            "project_id",
            "properties",
            "updated_at",
            "uuid",
            "workers",
        )
        self.assertEqual(collist, columns)

        datalist = (
            hardware_fakes.hardware_created_at,
            hardware_fakes.hardware_baremetal_type,
            hardware_fakes.hardware_name,
            hardware_fakes.hardware_project_id,
            {},
            hardware_fakes.hardware_updated_at,
            hardware_fakes.hardware_uuid,
            [],
        )
        self.assertEqual(datalist, tuple(data))


class TestHardwareList(TestHardware):
    def setUp(self):
        super().setUp()

        self.hardware_mock.list.return_value = list(
            hardware_fakes.FakeHardware.create_one_hardware()
        )

        self.cmd = hardware_cli.ListHardware(self.app, None)

    def test_hardware_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = []

        self.hardware_mock.list.assert_called_with(*args)


class TestHardwareCreate(TestHardware):
    def setUp(self):
        super().setUp()
        self.cmd = hardware_cli.CreateHardware(self.app, None)


class TestHardwareDelete(TestHardware):
    def setUp(self):
        super().setUp()
        self.cmd = hardware_cli.DeleteHardware(self.app, None)


class TestHardwareSet(TestHardware):
    def setUp(self):
        super().setUp()
        self.cmd = hardware_cli.UpdateHardware(self.app, None)


class TestHardwareSync(TestHardware):
    def setUp(self):
        super().setUp()
        self.cmd = hardware_cli.SyncHardware(self.app, None)



"""Define fields for Doni resources."""


class Resource(object):
    """Resource class.

    This class is used to manage the various fields that a resource (e.g.
    Chassis, Node, Port) contains.  An individual field consists of a
    'field_id' (key) and a 'label' (value).  The caller only provides the
    'field_ids' when instantiating the object.
    Ordering of the 'field_ids' will be preserved as specified by the caller.
    It also provides the ability to exclude some of these fields when they are
    being used for sorting.
    """

    FIELDS = {
        "name": "Name",
        "uuid": "UUID",
        "project_id": "Project ID",
        "hardware_type": "Hardware Type",
        "properties": "Properties",
        "created_at": "Created At",
        "updated_at": "Updated At",
        "workers": "Workers",
    }

    def __init__(self, field_ids, sort_excluded=None, override_labels=None):
        """Create a Resource object.

        :param field_ids:  A list of strings that the Resource object will
                           contain.  Each string must match an existing key in
                           FIELDS.
        :param sort_excluded: Optional. A list of strings that will not be used
                              for sorting.  Must be a subset of 'field_ids'.
        :param override_labels: Optional. A dictionary, where key is a field ID
                                and value is the label to be used. If
                                unspecified, uses the labels associated with
                                the fields from global FIELDS.
        :raises: ValueError if sort_excluded or override_labels contains values
                 not in field_ids
        """
        self._fields = tuple(field_ids)
        self._labels = tuple([self.FIELDS[x] for x in field_ids])

    @property
    def fields(self):
        return self._fields

    @property
    def labels(self):
        return self._labels


HARDWARE_RESOURCE = Resource(
    [
        "uuid",
        "name",
        "properties",
    ]
)


HARDWARE_DETAILED_RESOURCE = Resource(
    [
        "uuid",
        "name",
        "project_id",
        "hardware_type",
        "created_at",
        "updated_at",
        "properties",
    ]
)



import argparse
import json
import logging
from argparse import Namespace

import yaml
from cliff.columns import FormattableColumn
from keystoneauth1.exceptions import HttpError
from osc_lib import utils as osc_utils
from osc_lib.command import command

LOG = logging.getLogger(__name__)  # Get the logger of this module


class OutputFormat:
    columns = (
        "uuid",
        "name",
        "project_id",
        "hardware_type",
        "properties",
    )


class YamlColumn(FormattableColumn):
    def human_readable(self):
        return yaml.dump(self._value)


class HardwareSerializer(object):
    def serialize_hardware(self, hw_dict: "dict", columns: "list[str]"):
        return osc_utils.get_dict_properties(
            hw_dict,
            columns,
            formatters={
                "properties": YamlColumn,
                "workers": YamlColumn,
            },
        )


def conditional_action(ActionClass, condition_fn):
    class ConditionalAction(argparse.Action):
        def __call__(self, parser, namespace, values, *args, **kwargs):
            if condition_fn(namespace):
                return ActionClass(parser, namespace, values, *args, **kwargs)

    return ConditionalAction


class ExpandDotNotation(argparse.Action):
    """Set property based on dest.key."""

    def __call__(self, parser, namespace, values, option_string=None):
        group, dest = self.dest.split(".", 2)

        if not hasattr(namespace, group):
            setattr(namespace, group, {})
        getattr(namespace, group).update({dest: values})


class BaseParser(command.Command):
    """Base Parser for use with Doni commands.

    Behavior is to take arguments in the following forms, with later ones
    overriding earlier.
    - argparse "default" values
    - cli args
    - serialized json string (or json file for bulk)
    """

    needs_uuid = False

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("-d", "--dry-run", "--dry_run", action="store_true")
        parser.add_argument(
            "--json",
            default={},
            type=json.loads,
            help="Take args as json string. Values are overridden by other args.",
        )
        if self.needs_uuid:
            parser.add_argument(
                dest="uuid",
                metavar="<uuid>",
                help=("unique ID of hardware item"),
            )

        return parser

    def take_action(self, parsed_args: Namespace):
        """This allows setting arbitrary args via json input."""

        # Get dict of args without json string
        args_dict = vars(parsed_args)
        try:
            # convert json input to dict, remove from namespace
            json_dict = args_dict.get("json", {})
            # override parsed_args with json values
            args_dict.update(json_dict)
        except AttributeError:
            pass


class HardwarePatchCommand(BaseParser, HardwareSerializer):
    def get_patch(self, parsed_args):
        return []

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        hw_uuid = parsed_args.uuid

        patch = self.get_patch(parsed_args)

        if parsed_args.dry_run:
            LOG.warn(patch)
            return None

        if patch:
            try:
                LOG.debug(f"PATCH_BODY: {patch}")
                res = hw_client.update(hw_uuid, patch)
            except HttpError as ex:
                LOG.error(ex.response.text)
                raise ex
            else:
                return self.serialize_hardware(res.json(), list(OutputFormat.columns))
        else:
            LOG.info("No updates to send")



"""Define fake objects for Doni CLI tests."""

import copy
import datetime
import uuid
from unittest import mock

from openstackclient.tests.unit import fakes, utils

hardware_uuid = uuid.uuid4().hex
hardware_name = "fake name"
hardware_project_id = uuid.uuid4().hex
hardware_created_at = datetime.datetime.utcnow().isoformat()
hardware_updated_at = datetime.datetime.utcnow().isoformat()
hardware_baremetal_type = "baremetal"


class FakeHardware(object):
    @staticmethod
    def create_one_hardware(attrs=None):
        """Create a fake hw item."""
        attrs = attrs or {}

        hw_info = {
            "created_at": hardware_created_at,
            "hardware_type": hardware_baremetal_type,
            "name": hardware_name,
            "project_id": hardware_project_id,
            "properties": {},
            "updated_at": hardware_updated_at,
            "uuid": hardware_uuid,
            "workers": [],
        }

        hw_info.update(attrs)
        # TODO: client currently returns Dicts, not Resource objects.
        # hw_item = fakes.FakeResource(info=copy.deepcopy(hw_info), loaded=True)
        return hw_info


class FakeHardwareClient(object):
    def __init__(self, **kwargs) -> None:
        self.hardware = mock.Mock()
        self.hardware.resource_class = fakes.FakeResource(None, {})


class TestHardware(utils.TestCommand):
    def setUp(self):
        super(TestHardware, self).setUp()

        self.app.client_manager.auth_ref = mock.Mock(auth_token="TOKEN")
        self.app.client_manager.inventory = mock.Mock()


class FakeHardwareResource(fakes.FakeResource):
    def get_keys(self):
        return {"property": "value"}



"""Implements Doni command line interface."""

import argparse
from collections import namedtuple
import json
import logging
from argparse import FileType, Namespace

from keystoneauth1.exceptions import Conflict, HttpError
from osc_lib import utils as oscutils
from osc_lib.command import command

from doniclient.osc.common import (
    BaseParser,
    conditional_action,
    ExpandDotNotation,
    HardwarePatchCommand,
)
from doniclient.v1 import resource_fields as res_fields

LOG = logging.getLogger(__name__)  # Get the logger of this module

PropertyFlag = namedtuple("PropertyFlag", ["flag", "type", "default"])


class ListHardware(BaseParser, command.Lister):
    """List all hardware in the Doni database."""

    log = logging.getLogger(__name__ + ".ListHardware")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--all",
            help="List hardware from all owners. Requires admin rights.",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        """List all hw items in Doni."""
        columns = res_fields.HARDWARE_RESOURCE.fields
        labels = res_fields.HARDWARE_RESOURCE.labels

        hw_client = self.app.client_manager.inventory

        if parsed_args.all:
            data = hw_client.export()
        else:
            data = hw_client.list()

        return (
            labels,
            (
                oscutils.get_dict_properties(
                    s, columns, formatters={"Properties": oscutils.format_dict}
                )
                for s in data
            ),
        )


class GetHardware(BaseParser, command.ShowOne):
    """List specific hardware item in Doni."""

    needs_uuid = True

    def take_action(self, parsed_args):
        """List all hw items in Doni."""
        hw_client = self.app.client_manager.inventory
        try:
            data = hw_client.get_by_uuid(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex

        return self.dict2columns(data)


class DeleteHardware(BaseParser):
    """Delete specific hardware item in Doni."""

    needs_uuid = True

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        try:
            hw_client.delete(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex


class SyncHardware(BaseParser):
    """Sync specific hardware item in Doni."""

    needs_uuid = True

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        try:
            hw_client.sync(parsed_args.uuid)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex


def _add_prop_flag_group(parser, hardware_type, prop_flags):
    """Register a mapping of flags for corresponding hardware properties.

    Args:
        parser (argparse.Parser): the parent parser
        group_name (str): the name of the new argument grouping. This is just used
            to visually group the properties in the CLI usage text.
        prop_flags (dict): a mapping of property names to PropertyFlag objects
            containing information about how the flag should be displayed and parsed.
            Importantly, the property name MUST match some field on the ``properties``
            dict on the target hardware type.
    """
    # Only store this flag in the resulting args if the hardware_type is in effect.
    condition_fn = lambda args: args.hardware_type == hardware_type
    group = parser.add_argument_group(f"{hardware_type} properties")
    for prop_name, flag in prop_flags.items():
        group.add_argument(
            f"--{flag.flag}",
            type=flag.type,
            default=flag.default,
            metavar=f"<{flag.flag}>",
            dest=f"properties.{prop_name}",
            action=conditional_action(ExpandDotNotation, condition_fn),
        )


class CreateOrUpdateParser(BaseParser):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--name",
            help=(
                "Name of the hardware object. Best practice is to use a "
                "universally unique identifier, such has serial number or chassis ID. "
                "This will aid in disambiguating systems."
            ),
        )
        parser.add_argument(
            "--hardware_type",
            default="baremetal",
            help=(
                "Type of hardware object. Each type of hardware takes different "
                "properties."
            ),
        )

        _add_prop_flag_group(
            parser,
            "baremetal",
            {
                "mgmt_addr": PropertyFlag("mgmt_addr", str, None),
                "ipmi_username": PropertyFlag("ipmi_username", str, None),
                "ipmi_password": PropertyFlag("ipmi_password", str, None),
                "ipmi_terminal_port": PropertyFlag("ipmi_terminal_port", int, None),
                "baremetal_deploy_kernel_image": PropertyFlag(
                    "deploy_kernel", str, None
                ),
                "baremetal_deploy_ramdisk_image": PropertyFlag(
                    "deploy_ramdisk", str, None
                ),
                # FIXME(jason): why is the flag named ironic_?
                "baremetal_driver": PropertyFlag("ironic_driver", str, None),
                "baremetal_resource_class": PropertyFlag(
                    "resource_class", str, "baremetal"
                ),
                "baremetal_capabilities": PropertyFlag(
                    "capabilities", json.loads, None
                ),
                "cpu_arch": PropertyFlag("cpu_arch", str, "x86_64"),
                "node_type": PropertyFlag("blazar_node_type", str, None),
                "su_factor": PropertyFlag("blazar_su_factor", float, 1.0),
                "placement": PropertyFlag("placement", json.loads, None),
                "interfaces": PropertyFlag("interfaces", json.loads, {}),
            },
        )

        _add_prop_flag_group(
            parser,
            "device",
            {
                "machine_name": PropertyFlag("machine-name", str, None),
                "contact_email": PropertyFlag("contact-email", str, None),
                "channels": PropertyFlag("channels", json.loads, None),
                "application_credential_id": PropertyFlag(
                    "application-credential-id", str, None
                ),
                "application_credential_secret": PropertyFlag(
                    "application-credential-secret", str, None
                ),
            },
        )

        return parser


class CreateHardware(CreateOrUpdateParser):
    """Create a Hardware Object in Doni."""

    needs_uuid = False

    def get_parser(self, prog_name):
        return super().get_parser(prog_name)

    def take_action(self, parsed_args: Namespace):
        """Create new HW item."""
        # Call superclass action to parse input json
        super().take_action(parsed_args)

        hw_client = self.app.client_manager.inventory

        hw_type = parsed_args.hardware_type
        # FIXME(jason): 'device.balena' is really the only device type we use.
        # But, it's a bit awkard to have the user type this?
        if hw_type == "device":
            hw_type = "device.balena"

        body = {
            "name": parsed_args.name,
            "hardware_type": parsed_args.hardware_type,
            "properties": parsed_args.properties,
        }

        if parsed_args.dry_run:
            LOG.warn(parsed_args)
            LOG.warn(body)
            return

        try:
            data = hw_client.create(body)
        except HttpError as ex:
            LOG.error(ex.response.text)
            raise ex


class UpdateHardware(CreateOrUpdateParser, HardwarePatchCommand):
    """Update properties of existing hardware item."""

    needs_uuid = True

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        args_to_default = ("name", "properties")
        # Unset all defaults to avoid accidental changes
        for arg in parser._get_optional_actions():
            if arg.dest in args_to_default:
                arg.default = argparse.SUPPRESS

        return parser

    def get_patch(self, parsed_args):
        patch = []

        field_map = {
            "name": "name",
        }

        for key, val in field_map.items():
            arg = getattr(parsed_args, key, None)
            if arg:
                patch.append({"op": "add", "path": f"/{val}", "value": arg})

        try:
            for key, val in parsed_args.properties.items():
                patch.append({"op": "add", "path": f"/properties/{key}", "value": val})
        except AttributeError:
            pass

        return patch


class ImportHardware(BaseParser):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--skip_existing",
            help="continue if an item already exists, rather than exiting",
            action="store_true",
        )
        parser.add_argument("-f", "--file", help="JSON input file", type=FileType("r"))
        return parser

    def take_action(self, parsed_args):
        hw_client = self.app.client_manager.inventory
        with parsed_args.file as f:
            for item in json.load(f):
                if parsed_args.dry_run:
                    LOG.warn(item)
                else:
                    try:
                        data = hw_client.create(item)
                    except Conflict as ex:
                        LOG.error(ex.response.text)
                        if parsed_args.skip_existing:
                            continue
                        else:
                            raise ex
                    else:
                        LOG.debug(data)
